﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace TDP.BigFileFinder
{
    public partial class FrmMain : Form
    {
        private const long _BytesXKB = 1024;
        private const long _BytesXMB = 1024 * 1024;
        private const long _BytesXGB = 1024 * 1024 * 1024;
        private CancellationTokenSource _Cts;

        public FrmMain()
        {
            InitializeComponent();

            CmbFileSizeUnit.SelectedIndex = 2;
            TxtSize.Text = "1";
            TxtPath.Text = "C:\\";
        }

        private void MakeUISafeCall(Form form, Action act)
        {
            if (form.InvokeRequired)
            {
                form.Invoke(act);
            }
            else
            {
                act.Invoke();
            }
        }

        private List<MyFileInfo> ScanDirectory(string directoryName, long maxSizeBytes, CancellationTokenSource cts)
        {
            System.Diagnostics.Trace.WriteLine("Scanning directory: " + directoryName);

            MakeUISafeCall(this, () => LblStatus.Text = "Scanning directory: " + directoryName);

            List<MyFileInfo> Result = new List<MyFileInfo>();

            // Some directories could be inaccessible, like RecycleBin, so enclose the call 
            // to GetFiles in a try catch block and skip the remaining code if Files == null
            string[] Files = null;
            try
            {
                Files = Directory.GetFiles(directoryName);
            }
            catch {}
            
            if (Files != null)
            {
                foreach (string FileName in Files)
                {
                    System.Diagnostics.Trace.WriteLine("Checking file: " +  FileName);

                    if (cts.IsCancellationRequested)
                        throw new OperationCanceledException(cts.Token);

                    FileInfo CurrentFileInfo = new FileInfo(FileName);
                    if (CurrentFileInfo.Length > maxSizeBytes)
                    {
                        MyFileInfo Info = new MyFileInfo(FileName, CurrentFileInfo.Length);
                        Result.Add(Info);
                    }
                }

                string[] Directories = Directory.GetDirectories(directoryName);
                foreach (string directory in Directories)
                {
                    if (cts.IsCancellationRequested)
                        throw new OperationCanceledException(cts.Token);

                    List<MyFileInfo> SubDirResult = ScanDirectory(directory, maxSizeBytes, cts);

                    if (SubDirResult.Count > 0)
                        Result.InsertRange(Result.Count, SubDirResult);
                }
            }

            return Result;
        }

        private Task<List<MyFileInfo>> ScanDirectoryAsync(string directoryName, long maxSizeBytes, CancellationTokenSource cts)
        {
            Task<List<MyFileInfo>> T = new Task<List<MyFileInfo>>(() =>
            {
                List<MyFileInfo> Result = ScanDirectory(directoryName, maxSizeBytes, cts);

                return Result;
            }, cts.Token);
            T.Start();
            return T;
        }

        private void MnuExit_Click(object sender, EventArgs e)
        {
            Close();
        }

        private async void BtnStart_Click(object sender, EventArgs e)
        {
            string StartPath = TxtPath.Text;
            if (!StartPath.EndsWith("/"))
                StartPath += "/";
            
            if (!Directory.Exists(StartPath))
            {
                MessageBox.Show($"The path '{StartPath}' doesn't exist.", "TDP Big File Finder", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                return;
            }

            long MaxFileSize;
            bool IsValidLong = long.TryParse(TxtSize.Text, out MaxFileSize);
            if (!IsValidLong)
            {
                MessageBox.Show($"The field 'File size' doesn't contain a vaild number.", "TDP Big File Finder", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                return;
            }

            switch (CmbFileSizeUnit.Text)
            {
                case "KB":
                    MaxFileSize *= _BytesXKB;
                    break;
                case "MB":
                    MaxFileSize *= _BytesXMB;
                    break;
                case "GB":
                    MaxFileSize *= _BytesXGB;
                    break;
            }

            _Cts = new CancellationTokenSource();

            BtnStart.Enabled = false;
            BtnStop.Enabled = true;

            try
            {             
                List<MyFileInfo> Result = await ScanDirectoryAsync(TxtPath.Text, MaxFileSize, _Cts);

                if (Result.Count == 0)
                    MessageBox.Show("Search completed, but no files were found.", "TDP Big File Finder", MessageBoxButtons.OK, MessageBoxIcon.Information);
                else
                {
                    MessageBox.Show("Search completed.", "TDP Big File Finder", MessageBoxButtons.OK, MessageBoxIcon.Information);
                    DgResults.DataSource = Result.OrderByDescending(r => r.Size).ToList();
                }

                BtnStart.Enabled = true;
                BtnStop.Enabled = false;
            }
            catch (OperationCanceledException)
            {
                BtnStart.Enabled = true;
                BtnStop.Enabled = false;
                LblStatus.Text = "Operation stopped by the user.";
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "TDP Big File Finder", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void BtnStop_Click(object sender, EventArgs e)
        {
            if (_Cts != null)
            {
                _Cts.Cancel();
                _Cts = null;
            }
        }
    }
}
